#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/args.py

#
# options used
# -b --backend
# -d --default
# -l --layer
# -o --output-directory
# -p --prefix
# # -s --seed-devices
# -u --user-data-file
# -v --version
# --api-port (deprecated ?)
# --case
# --check-user-data-only
# --dont-compare
# --filter-customers
# --filter-sites
# --include-l3-hosts
# --keep
# --log-file
# --log-level
# --log-to-stdout
# --min-age
# --pre-fetch
# --quiet
# --remove-domain
# --skip-l3-if
# --skip-l3-ip
# --time-format


from argparse import (
    ArgumentParser,
    Namespace as arg_Namespace,
    RawTextHelpFormatter,
)
from pathlib import Path

from lib.constants import (
    HOME_URL,
    MIN_CDP_VERSION,
    MIN_LINUX_IP_ADDRESSES,
    MIN_SNMP_IP_ADDRESSES,
    MIN_WINDOWS_IP_ADDRESSES,
    MIN_LLDP_VERSION,
    NVDCT_VERSION,
    SCRIPT,
    TIME_FORMAT_ARGPARSER,
    USER_DATA_FILE,
)
from lib.utils import ExitCodes


def parse_arguments() -> arg_Namespace:
    parser = ArgumentParser(
        prog='nvdct.py',
        description=(
            'This script creates the topology data file needed for the Checkmk "network_visualization"\n'  # noqa: E501
            'plugin by Andreas Boesl and schnetz. For more information see\n'
            'the announcement from schnetz: '
            'https://forum.checkmk.com/t/network-visualization/41680\n'
            'and the plugin on the Exchange: '
            'https://exchange.checkmk.com/p/network-visualization .\n'
            '\n'
            'The required inventory data can be created with my inventory plugins:\n'
            'CDP: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_cdp_cache\n'  # noqa: E501
            'LLDP: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lldp_cache\n'  # noqa: E501
            'L3v4: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ip_address\n'  # noqa: E501
            '\n'
            f'\nVersion: {NVDCT_VERSION} | Written by: thl-cmk\n'
            f'for more information see: {HOME_URL}'
        ),
        formatter_class=RawTextHelpFormatter,
        epilog='Exit codes:\n'
               f' {ExitCodes.OK.value} - No error\n'
               f' {ExitCodes.BAD_OPTION_LIST.value} - Bad options list\n'
               f' {ExitCodes.BACKEND_NOT_IMPLEMENTED.value} - Backend not implemented\n'
               f' {ExitCodes.AUTOMATION_SECRET_NOT_FOUND.value} - Automation secret not found\n'
               '\nUsage:\n'
               f'{SCRIPT} -u ~/local/bin/nvdct/conf/my_{USER_DATA_FILE} \n\n'
    )

    parser.add_argument(
        '-b', '--backend',
        choices=['LIVESTATUS', 'MULTISITE', 'RESTAPI'],
        # default='MULTISITE',
        help='Backend used to retrieve the topology data\n'
             ' - LIVESTATUS : fetches data via local Livestatus (local site only)\n'
             ' - MULTISITE  : like LIVESTATUS but for distributed environments (default)\n'
             ' - RESTAPI    : uses the CMK REST API.',
    )
    parser.add_argument(
        '-d', '--default', action='store_const', const=True,  # default=False,
        help='Set the created topology data as default. Will be created automatically\n'
             'if it doesn\'t exists.',
    )
    parser.add_argument(
        '-o', '--output-directory', type=str,
        help='Directory name where to save the topology data.\n'
             'I.e.: my_topology. Default is the actual date/time\n'
             'in "--time-format" format.\n'
             'NOTE: the directory is a sub directory under "~/var/check_mk/topology/data/"\n',
    )
    # parser.add_argument(
    #     '-s', '--seed-devices', type=str, nargs='+',
    #     help=f'List of devices to start the topology discovery from.\n'
    #          f'I.e. {SAMPLE_SEEDS}',
    # )
    parser.add_argument(
        '-p', '--prefix', type=str,
        help='Prepends each host with the prefix. (Needs testing)\n'
    )
    parser.add_argument(
        '-l', '--layers',
        nargs='+',
        choices=['CDP', 'CUSTOM', 'LLDP', 'STATIC', 'L3v4'],
        # default=['CDP'],
        help=(
            f' - CDP   : needs inv_cdp_cache package at least in version {MIN_CDP_VERSION}\n'
            f' - LLDP  : needs inv_lldp_cache package at least in version {MIN_LLDP_VERSION}\n'
            f' - L3v4  : needs inv_ip_address package at least in version {MIN_SNMP_IP_ADDRESSES} for SNMP based hosts\n'
            f'           for Linux based hosts inv_lnx_ip_if in version {MIN_LINUX_IP_ADDRESSES}\n'
            f'           for Windows based hosts inv_win_ip_if in version {MIN_WINDOWS_IP_ADDRESSES}\n'
            f' - STATIC: creates a topology base on the "STATIC_CONNECTIONS" in the toml file\n'
            f' - CUSTOM: (deprecated)\n'
        )
    )
    parser.add_argument(
        '-u', '--user-data-file', type=str,
        help='Set the name uf the user provided data file\n'
             f'Default is ~/local/bin/nvdct/conf/{USER_DATA_FILE}\n',
    )
    parser.add_argument(
        '-v', '--version', action='version',
        version=f'{Path(SCRIPT).name} version: {NVDCT_VERSION}',
        help='Print version of this script and exit',
    )
    parser.add_argument(
        '--api-port', type=int,  # default=False,
        help='TCP Port to access the REST API. Default is 80. NVDCT will try to automatically\n'
             'detect the site apache port.',
    )
    parser.add_argument(
        '--case',
        choices=['LOWER', 'UPPER'],
        # default='NONE',
        help='Change neighbour name to all lower/upper case',
    )
    parser.add_argument(
        '--check-user-data-only', action='store_const', const=True,  # default=False,
        help=f'Only tries to read/parse the user data from {USER_DATA_FILE} and exits.',
    )
    parser.add_argument(
        '--log-file', type=str,
        help='Set the log file. Default is ~/var/log/nvdct.log\n',
    )
    parser.add_argument(
        '--log-level',
        # nargs='+',
        choices=['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'OFF'],
        # default='WARNING',
        help='Sets the log level. The default is "WARNING"\n'
    )
    parser.add_argument(
        '--log-to-stdout', action='store_const', const=True,  # default=False,
        help='Send log to stdout.',
    )
    parser.add_argument(
        '--dont-compare', action='store_const', const=True,  # default=False,
        help='Do not compare the actual topology data with the default topology\n'
             'data. By default, the actual topology is compared with the default\n'
             'topology. If the data matches, the actual topology is not saved.\n'
             'So, if you run this tool in a cron job, a new topology will be\n'
             'created only if there was a change, unless you use "--dont-compare".'
    )
    parser.add_argument(
        '--filter-customers',
        choices=['INCLUDE', 'EXCLUDE'],
        # default='INCLUDE',
        help='INCLUDE/EXCLUDE customer list from config file.\n'
             'Note: MULTISITE backend only.',
    )
    parser.add_argument(
        '--filter-sites',
        choices=['INCLUDE', 'EXCLUDE'],
        # default='INCLUDE',
        help='INCLUDE/EXCLUDE site list from config file.\n'
             'Note: MULTISITE backend only.',
    )
    parser.add_argument(
        '--include-l3-hosts', action='store_const', const=True,  # default=False,
        help='Include hosts (single IP objects) in layer 3 topology',
    )
    parser.add_argument(
        '--remove-domain', action='store_const', const=True,  # default=False,
        help='Remove the domain name from the neighbor name',
    )
    parser.add_argument(
        '--keep', type=int,
        help='Number of topologies to keep. The oldest topologies above keep\n'
             'max will be deleted.\n'
             'NOTE: The default topologies will be always kept.\n'
    )
    parser.add_argument(
        '--min-age', type=int,
        help='The minimum number of days before a topology is deleted by "--keep".'
    )
    parser.add_argument(
        '--pre-fetch', action='store_const', const=True,  # default=False,
        help='Try to fetch host data, with less API calls. Can improve RESTAPI backend\n'
             'performance',
    )
    parser.add_argument(
        '--quiet', action='store_const', const=True,  # default=False,
        help='Suppress output to stdtout',
    )
    parser.add_argument(
        '--skip-l3-if', action='store_const', const=True,  # default=False,
        help='Skip interface in layer 3 topology',
    )
    parser.add_argument(
        '--skip-l3-ip', action='store_const', const=True,  # default=False,
        help='Skip ip-address in layer 3 topology',
    )
    parser.add_argument(
        '--time-format', type=str,
        help=f'Format string to render the time. (default: {TIME_FORMAT_ARGPARSER})',
    )

    return parser.parse_args()
