#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2024-06-09
# File  : lib/topologies.py

from collections.abc import Mapping, Sequence
from ipaddress import IPv4Address, IPv4Network
from typing import Dict, List, Tuple
from re import sub as re_sub

from lib.backends import (
    CacheItems,
    HostCache,
)
from lib.constants import (
    CACHE_INTERFACES_DATA,
    HOST_LABEL_L3V4_HOSTS,
    HOST_LABEL_L3V4_ROUTER,
    LOGGER,
    PATH_INTERFACES,
    PATH_L3v4,
)
from lib.settings import (
    Emblems,
    StaticConnection,
    Thickness,
    Wildcard,
)
from lib.utils import (
    InventoryColumns,
    Ipv4Info,
    is_valid_hostname,
)


class NvObjects:
    def __init__(self) -> None:
        self.nv_objects: Dict[str, any] = {}
        self.host_count: int = 0
        self.host_list: List[str] = []

    def add_host(
        self,
        host: str,
        host_cache: HostCache,
        emblem: str | None = None
    ) -> None:
        if host not in self.nv_objects:
            self.host_count += 1
            self.host_list.append(host)
            link: Dict = {}
            metadata: Dict = {}
            # LOGGER.debug(f'host: {host}, {host_cache.host_exists(host=host)}')
            if host_cache.host_exists(host=host) is True:
                LOGGER.debug(f'host: {host} exists')
                link = {'core': host}
            else:
                if emblem is not None:
                    metadata = {
                        'images': {
                            'emblem': emblem,  # node image
                            # 'icon': 'icon_tick',  # status image, top left from emblem
                        },
                        **({'tooltip': {
                            'quickinfo': [{'name': 'Host node', 'value': host}]
                        }} if not link else {})
                    }

            self.nv_objects[host] = {
                'name': host,
                'link': link,
                'metadata': metadata,
            }
            LOGGER.debug(f'host: {host}, link: {link}, metadata: {metadata}')

    def add_service(
        self,
        host: str,
        service: str,
        host_cache: HostCache,
        emblem: str | None = None,
        metadata: Dict | None = None,
        name: str | None = None,
    ) -> None:
        if metadata is None:
            metadata = {}
        if name is None:
            name = service

        self.add_host(host=host, host_cache=host_cache)
        service_object = f'{service}@{host}'
        if service_object not in self.nv_objects:
            link: Dict = {}
            if host_cache.host_exists(host=host):
                link = {'core': [host, service]}
            elif emblem is not None:
                metadata.update({
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    **({'tooltip': {'quickinfo': [{'name': 'Service node', 'value': service}]}} if not link else {})
                })

            self.nv_objects[service_object] = {
                'name': name,
                'link': link,
                'metadata': metadata,
            }
        elif metadata is not {}:
            self.nv_objects[service_object]['metadata'].update(metadata)


    def add_interface(
        self,
        host: str,
        service: str,
        host_cache: HostCache,
        emblem: str | None = None,
        metadata: Dict | None = None,
        name: str | None = None,
        item: str | None = None,
    ) -> None:
        if metadata is None:
            metadata = {}
        if name is None:
            name = service
        speed = None

        self.add_host(host=host, host_cache=host_cache)
        service_object = f'{service}@{host}'
        if service_object not in self.nv_objects:
            link: Dict = {}
            if item is None:
                item = get_service_by_interface(host, service.lstrip('0'), host_cache)
            if item and host_cache.host_exists(host=host):
                service_long = f'Interface {item}'
                link = {'core': [host, service_long]}
                if op_data := get_operational_interface_data(host, item, host_cache):
                    metadata.update(op_data)
                    speed = op_data.get('op_speed_int')
            elif emblem is not None:
                metadata.update({
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    **({'tooltip': {'quickinfo': [
                        {'name': 'Service node', 'value': item}
                    ]}} if not link else {})
                })
            if isinstance(speed, int):
                metadata.update({'native_speed': speed})

            self.nv_objects[service_object] = {
                'name': name,
                'link': link,
                'metadata': metadata,
            }
        elif metadata is not {}:
            self.nv_objects[service_object]['metadata'].update(metadata)

    def add_ipv4_address(
        self,
        host: str,
        ipv4_address: str,
        emblem: str,
        interface: str | None,
    ) -> None:
        if interface is not None:
            service_object = f'{ipv4_address}@{interface}@{host}'
        else:
            service_object = f'{ipv4_address}@{host}'

        if service_object not in self.nv_objects:
            self.nv_objects[service_object] = {
                'name': ipv4_address,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,
                        # 'icon': 'icon_tick',
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Address', 'value': ipv4_address},
                        ]
                    }
                }
            }

    def add_ipv4_network(self, network: str, emblem: str, ) -> None:
        if network not in self.nv_objects:
            self.nv_objects[network] = {
                'name': network,
                'link': {},
                'metadata': {
                    'images': {
                        'emblem': emblem,  # node image
                        # 'icon': 'icon_tick',  # status image, top left from emblem
                    },
                    'tooltip': {
                        'quickinfo': [
                            {'name': 'IP-Network', 'value': network},
                        ]
                    }
                }
            }

    def add_tooltip_quickinfo(self, nv_object: str, name: str, value: str) -> Dict:
        metadata = self.nv_objects[nv_object]['metadata']
        if metadata.get('tooltip') is None:
            metadata['tooltip'] = {}
        if metadata['tooltip'].get('quickinfo') is None:
            metadata['tooltip']['quickinfo'] = []

        metadata['tooltip']['quickinfo'].append({
            'name': name,
            'value': value,
        })

        return metadata

    def add_icon_to_object(self, nv_object: str, icon: str) -> None:
        if nv_object in self.nv_objects:
            if not self.nv_objects[nv_object]['metadata'].get('images'):
                self.nv_objects[nv_object]['metadata']['images'] = {}
            self.nv_objects[nv_object]['metadata']['images']['icon'] = icon


class NvConnections:
    def __init__(self) -> None:
        self.nv_connections: List = []

    def add_connection(self, left: str, right: str):
        connection = [left, right]
        connection.sort()
        if [connection] not in self.nv_connections:
            self.nv_connections.append([connection])

    def add_meta_data_to_connections(
        self,
        nv_objects: NvObjects,
        speed_map: Sequence[Thickness],
    ):
        for connection in self.nv_connections:
            warning = False
            left, right = connection[0]
            metadata: Dict[str, Dict] = {
                'line_config': {
                    'css_styles': {
                        'stroke-dasharray': 'unset'
                    }
                }
            }

            left_speed = nv_objects.nv_objects[left].get('metadata', {}).get('op_speed_int')
            right_speed = nv_objects.nv_objects[right].get('metadata', {}).get('op_speed_int')

            left_speed_str = nv_objects.nv_objects[left].get('metadata', {}).get('op_sped_str')
            right_speed_str = nv_objects.nv_objects[right].get('metadata', {}).get('op_sped_str')

            left_duplex = nv_objects.nv_objects[left].get('metadata', {}).get('duplex')
            right_duplex = nv_objects.nv_objects[right].get('metadata', {}).get('duplex')

            left_native_vlan = nv_objects.nv_objects[left].get('metadata', {}).get('native_vlan')
            right_native_vlan = nv_objects.nv_objects[right].get('metadata', {}).get('native_vlan')


            if right_speed and left_speed:
                right_thickness = map_speed_to_thickness(right_speed, speed_map)
                # left_thickness = map_speed_to_thickness(left_speed, speed_map)
                metadata['line_config']['thickness'] = right_thickness

                if right_speed != left_speed:
                    warning = True
                    metadata = add_tooltip_html(
                        metadata, 'Speed', left, left_speed_str, right, right_speed_str
                    )
                    # metadata = add_tooltip_quickinfo(metadata, 'Warning', 'Speed mismatch')
                    # metadata = add_tooltip_quickinfo(metadata, left, left_speed_str)
                    # metadata = add_tooltip_quickinfo(metadata, right, right_speed_str)

                    LOGGER.warning(
                        f'Connection with speed mismatch: {left} (speed: {left_speed_str})'
                        f'<->{right} (speed: {right_speed_str})'
                    )

            # for duplex/native vlan it might be a good idea to change left/right
            # value as they are reported by CDP as neighbour states
            if left_duplex and right_duplex:
                if left_duplex != right_duplex:
                    warning = True

                    metadata = add_tooltip_html(
                        metadata, 'Duplex', left, left_duplex, right, right_duplex
                    )

                    LOGGER.warning(
                        f'Connection with duplex mismatch: {left} (duplex: {left_duplex})'
                        f'<->{right} (duplex: {right_duplex})'
                    )
            if left_native_vlan and right_native_vlan:
                if left_native_vlan != right_native_vlan:
                    warning = True

                    metadata = add_tooltip_html(
                        metadata, 'Native<br>VLAN', left, left_native_vlan, right, right_native_vlan
                    )

                    LOGGER.warning(
                        f'Connection with native vlan mismatch: '
                        f'{left} (vlan: {left_native_vlan})<->{right} (vlan: {right_native_vlan})'
                    )
            if warning:
                metadata['line_config'].update({
                        'color': 'red',
                        'thickness': 5,
                })
                metadata['line_config']['css_styles']['stroke-dasharray'] = '10 5'
                nv_objects.add_icon_to_object(left, 'icon_warning')
                nv_objects.add_icon_to_object(right, 'icon_warning')
                close_tooltip_html(metadata)
            connection.append(metadata)


def map_speed_to_thickness(speed_to_map: int, speed_map: Sequence[Thickness]) -> int:
    thickness: int = 1  # use in case of empty MAP_SPEED_TO_THICKNESS
    for speed, thickness in speed_map:
        if speed_to_map <= speed:
            break
    return thickness


def get_operational_interface_data(
    host: str,
    item: str,
    host_cache: HostCache,

) -> Dict[str, str | int] | None:
    unit_to_bits_per_second = {
        'Bit/s': 1,
        'kBit/s': 1000,
        'Kbps': 1000,
        'MBit/s': 1000000,
        'Mbps': 1000000,
        'GBit/s': 1000000000,
        'Gbps': 1000000000,
    }

    # get dict of interfaces with the item as key
    interface_data: Dict[str, any] | None = host_cache.get_data(
        host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
    )
    try:
        raw_opdata = interface_data[item]['long_plugin_output']
    except (KeyError, TypeError):
        return None

    if raw_opdata:
        opdata: Dict[str, str | int] = {}
        for _entry in raw_opdata:
            try:
                key, value = _entry.split(': ', 1)  # split only at the first colon
            except ValueError:
                continue
            value = value.strip(' ')
            match key:
                case 'MAC':
                    if len(value) == 17:  # valid MAC: 6C:DD:30:DD:51:8B'
                        opdata['mac'] = value
                case 'Speed':
                    try:           # *_ -> ignore rest of string, i.e: (expected: 1 GBit/s)WARN
                        speed, unit, *_ = value.split(' ')
                    except ValueError:
                        pass
                    else:
                        opdata['op_sped_str'] = f'{speed} {unit}'
                        opdata['op_speed_int'] = int(float(speed) * unit_to_bits_per_second[unit])

        return opdata
    return None


def get_service_by_interface(host: str, interface: str, host_cache: HostCache) -> str | None:
    """
    Returns:
        Tuple of interface item
    """

    _alternate_if_name = {
        'ethernet': 'eth',
        # 'fastethernet': 'Fa',
        # 'gigabitethernet': 'gi',
        # 'tengigabitethernet': 'te',
        # 'fortygigabitethernet': 'Fo',
        # 'hundredgigabitethernet': 'Hu',
        # 'management': 'Ma',
    }

    def _get_short_if_name(interface_: str) -> str:
        """
        returns short interface name from long interface name
        interface: is the long interface name
        :type interface_: str
        """
        if not interface_:
            return interface_
        for interface_prefix in _alternate_if_name.keys():
            if interface_.lower().startswith(interface_prefix.lower()):
                interface_short = _alternate_if_name[interface_prefix]
                return interface_.lower().replace(interface_prefix.lower(), interface_short, 1)
        return interface_

    # try to find the item for an interface
    def _match_entry_with_item(_entry: Mapping[str, str], services: Sequence[str]) -> str | None:
        values = [
            _entry.get('name'.strip()),
            _entry.get('description'.strip()),
            _entry.get('alias').strip()
        ]
        for value in values:
            if value in services:
                return value

        index = str(_entry.get('index'))

        # try alias+index
        alias_index = str(_entry.get('alias')).strip() + ' ' + index
        if alias_index in services:
            LOGGER.info(f'match found by alias-index|{_entry}|{alias_index}|')
            return alias_index

        # try descrption+index
        description_index = str(_entry.get('description')).strip() + ' ' + index
        if description_index in services:
            LOGGER.info(f'match found by alias-index|{_entry}|{description_index}|')
            return description_index

        # for index try with padding
        pad_services: List[str] = [x for x in services if x.isdigit()]
        if pad_services:
            max_pad = len(max(pad_services, key=len)) + 1
            min_pad = len(min(pad_services, key=len))
            for i in range(min_pad, max_pad):
                index_padded = f'{index:0>{i}}'
                if index_padded in pad_services:
                    return index_padded
                # still not found try values + index
                for value in values:
                    if f'{value} {index_padded}' in services:
                        return f'{value} {index_padded}'

        LOGGER.warning(f'no match found |{_entry}|{services}|')
        return None

    # empty host/neighbour should never happen here
    if not host:
        LOGGER.warning(f'no host name |{host}|')
        return None

    # get dict of interfaces with the item as key
    interface_data: Mapping[str, Mapping[str, object]] = host_cache.get_data(
        host=host, item=CacheItems.interfaces, path=CACHE_INTERFACES_DATA
    )
    if not interface_data:
        LOGGER.warning(f'got no interface data for: {host}')
        return None

    # try to find the interface in the host interface inventory list
    inventory = host_cache.get_data(
        host=host, item=CacheItems.inventory, path=PATH_INTERFACES
    )
    if not inventory:
        LOGGER.warning(f'no interface inventory for host: {host}')
        return None

    interface_items: Sequence[str] = list(interface_data.keys())

    # the easy case
    if interface in interface_items:
        return interface

    for _entry in inventory:
        if interface in [
            _entry.get('name'),
            _entry.get('description'),
            _entry.get('alias'),
            str(_entry.get('index')),
            _entry.get('phys_address'),
        ]:
            return _match_entry_with_item(_entry, interface_items)
        elif f'1:{interface}' == _entry.get('name'):  # Extreme non stack:
            return _match_entry_with_item(_entry, interface_items)
        elif _entry.get('name') is not None and _get_short_if_name(
                _entry.get('name')) == str(interface).lower():  # Cisco NXOS
            return _match_entry_with_item(_entry, interface_items)

    LOGGER.warning(msg=f'Device: {host}: service for interface |{interface}| not found')


def add_tooltip_html(
        metadata: Dict,
        type_: str,
        left: str,
        left_value: str,
        right: str,
        right_value: str,
) -> Dict:
    css_class = 'class=mismatch'
    if metadata.get('tooltip') is None:
        metadata['tooltip'] = {}
    if metadata['tooltip'].get('html') is None:
        css_style = (
            '<style>'
                'div.mismatch {'
                    'background-color: rgba(70,70,70,.1);'
                    'border-radius: 5px;'
                    'padding: 5px;'
                '}'
                'p.mismatch {'
                    'text-align: center;'
                '}'
                'table.mismatch {'
                    'text-align: left;'
                    'border-radius: 5px;'
                '}'
                'tr.mismatch {'
                '}'
                'tr.mismatch:nth-child(even) {'
                    'background-color: rgba(100,100,100,.3);'

                '}'
                'tr.mismatch:nth-child(odd) {'
                    'background-color: rgba(100,100,100,.2);'

                '}'
                'th.mismatch {'
                    'background-color: rgba(120,120,120,.3);'
                    'padding: 5px;'   # inside the element
                    # 'margin: 5px;'  # outside of the lement
                '}'
                'td.mismatch {'
                    'padding: 5px;'
                '}'
            '</style>'
        )
        header = (
            '<thead>'
                '<tr>'
                    f'<th {css_class}>Node</th>'
                    f'<th {css_class}>{left}</th>'
                    f'<th {css_class}>{right}</th>'
                '</tr>'
            '</thead>'
        )
        metadata['tooltip']['html'] = (
            f'{css_style}'
            f'<div {css_class}>'
                f'<p {css_class}>WARNING: Mismatch found!</p>'
                f'<table {css_class}>'
                    f'{header}'
                    '<tbody>'
        )

    metadata['tooltip']['html'] += (
        f'<tr {css_class}>'
            f'<td {css_class}>{type_}</td>'
            f'<td {css_class}>{left_value}</td>'
            f'<td {css_class}>{right_value}</td>'
        '</tr>'
    )

    return metadata


def close_tooltip_html(metadata: Dict) -> Dict:
    metadata['tooltip']['html'] += '</table></div>'
    return metadata


def get_network_summary(ipv4_address: str, summarize: Sequence[IPv4Network]) -> str | None:
    for network in summarize:
        if IPv4Network(ipv4_address).subnet_of(network):
            return network.exploded
    return None


def is_ignore_ipv4(ip_address: str, ignore_ips: Sequence[IPv4Network]) -> bool:
    for ip in ignore_ips:
        if IPv4Network(ip_address).subnet_of(ip):
            LOGGER.info(f'IP address {ip_address} is in ignore list -> ({ip})')
            return True
    return False


def is_ignore_wildcard(ip_address: str, ignore_wildcard: Sequence[Wildcard]) -> bool:
    int_ip_address = int(IPv4Address(ip_address))
    for wildcard in ignore_wildcard:
        if int_ip_address & wildcard.int_wildcard == wildcard.bit_pattern:
            LOGGER.info(
                f'IP address {ip_address} matches ignore wildcard '
                f'list ({wildcard.ip_address}/{wildcard.wildcard})'
            )
            return True
    return False


def get_list_of_devices(data: Mapping) -> List[str]:
    devices: List[str] = []
    for connection in data.values():
        devices.append(connection[0])
    return list(set(devices))


def create_static_connections(
        connections_: Sequence[StaticConnection],
        emblems: Emblems,
        host_cache: HostCache,
        nv_connections: NvConnections,
        nv_objects: NvObjects,
):
    for connection in connections_:
        LOGGER.info(msg=f'connection: {connection}')
        nv_objects.add_host(
            host=connection.right_host,
            host_cache=host_cache,
            emblem=emblems.host_node
        )
        nv_objects.add_host(
            host=connection.left_host,
            host_cache=host_cache,
            emblem=emblems.host_node
        )
        if connection.right_service:
            nv_objects.add_service(
                host=connection.right_host,
                host_cache=host_cache,
                emblem=emblems.service_node,
                service=connection.right_service
            )
            nv_connections.add_connection(
                left=connection.right_host,
                right=f'{connection.right_service}@{connection.right_host}',
            )

        if connection.left_service:
            nv_objects.add_service(
                host=connection.left_host,
                host_cache=host_cache,
                emblem=emblems.service_node,
                service=connection.left_service
            )
            nv_connections.add_connection(
                left=connection.left_host,
                right=f'{connection.left_service}@{connection.left_host}',
            )

        if connection.right_service and connection.left_service:
            nv_connections.add_connection(
                left=f'{connection.right_service}@{connection.right_host}',
                right=f'{connection.left_service}@{connection.left_host}',
            )
        elif connection.right_service:  # connect right_service with left_host
            nv_connections.add_connection(
                left=f'{connection.right_service}@{connection.right_host}',
                right=f'{connection.left_host}',
            )
        elif connection.left_service:  # connect left_service with right_host
            nv_connections.add_connection(
                left=f'{connection.right_host}',
                right=f'{connection.left_service}@{connection.left_host}',
            )
        else:  # connect right_host with left_host
            nv_connections.add_connection(
                left=f'{connection.right_host}',
                right=f'{connection.left_host}',
            )


def create_l2_device_from_inv(
        case: str,
        host: str,
        host_cache: HostCache,
        inv_columns: InventoryColumns,
        inv_data: Sequence[Mapping[str, str]],
        l2_drop_hosts: List,
        l2_host_map: Dict[str, str],
        l2_neighbour_replace_regex: List[Tuple[str, str]] | None,
        nv_connections: NvConnections,
        nv_objects: NvObjects,
        prefix: str,
        remove_domain: bool,
) -> None:
    for topo_neighbour in inv_data:
        # check if required data are not empty
        if not (neighbour := topo_neighbour.get(inv_columns.neighbour)):
            LOGGER.warning(f'incomplete data, neighbour missing {topo_neighbour}')
            continue
        if not (raw_local_port := topo_neighbour.get(inv_columns.local_port)):
            LOGGER.warning(f'incomplete data, local port missing {topo_neighbour}')
            continue
        if not (raw_neighbour_port := topo_neighbour.get(inv_columns.neighbour_port)):
            LOGGER.warning(f'incomplete data, neighbour port missing {topo_neighbour}')
            continue

        # drop neighbour before domain split
        if neighbour in l2_drop_hosts:
            LOGGER.info(msg=f'drop neighbour: {neighbour}')
            continue

        if l2_neighbour_replace_regex:
            for re_str, replace_str in l2_neighbour_replace_regex:
                re_neighbour = re_sub(re_str, replace_str, neighbour)
                if re_neighbour != neighbour:
                    LOGGER.info(f'regex changed Neighbor |{neighbour}| to |{re_neighbour}|')
                    neighbour = re_neighbour
                if not neighbour:
                    LOGGER.info(f'Neighbour removed by regex (|{neighbour}|, |{re_str}|, |{replace_str}|)')
                    break
            if not neighbour:
                continue

        if remove_domain:
            neighbour = neighbour.split('.')[0]

        # drop neighbour after domain split
        if neighbour in l2_drop_hosts:
            LOGGER.info(msg=f'drop neighbour: {neighbour}')
            continue

        if case == 'UPPER':
            neighbour = neighbour.upper()
            LOGGER.debug(f'Changed neighbour to upper case: {neighbour}')
        elif case == 'LOWER':
            neighbour = neighbour.lower()
            LOGGER.debug(f'Changed neighbour to lower case: {neighbour}')

        if prefix:
            neighbour = f'{prefix}{neighbour}'
        # rewrite neighbour if inventory neighbour and checkmk host don't match
        if neighbour in l2_host_map.keys():
            neighbour = l2_host_map[neighbour]

        # getting/checking interfaces
        local_port = get_service_by_interface(host, raw_local_port, host_cache)
        if not local_port:
            local_port = raw_local_port
            LOGGER.warning(msg=f'service not found: host: {host}, raw_local_port: {raw_local_port}')
        elif local_port != raw_local_port:
            # local_port = raw_local_port  # don't reset local_port
            LOGGER.info(
                msg=f'host: {host}, raw_local_port: {raw_local_port} -> local_port: {local_port}'
            )

        neighbour_port = get_service_by_interface(neighbour, raw_neighbour_port, host_cache)
        if not neighbour_port:
            neighbour_port = raw_neighbour_port
            LOGGER.warning(
                msg=f'service not found: neighbour: {neighbour}, '
                    f'raw_neighbour_port: {raw_neighbour_port}'
            )
        elif neighbour_port != raw_neighbour_port:
            # neighbour_port = raw_neighbour_port  # don't reset neighbour_port
            LOGGER.info(
                msg=f'neighbour: {neighbour}, raw_neighbour_port {raw_neighbour_port} '
                    f'-> neighbour_port {neighbour_port}'
            )

        metadata = {
            'duplex': topo_neighbour.get('duplex'),
            'native_vlan': topo_neighbour.get('native_vlan'),
        }

        nv_objects.add_host(host=host, host_cache=host_cache)
        nv_objects.add_host(host=neighbour, host_cache=host_cache)
        nv_objects.add_interface(
            host=str(host),
            service=str(local_port),
            host_cache=host_cache,
            metadata=metadata,
            name=str(raw_local_port),
            item=str(local_port)
        )
        nv_objects.add_interface(
            host=str(neighbour),
            service=str(neighbour_port),
            host_cache=host_cache,
            name=str(raw_neighbour_port),
            item=str(neighbour_port)
        )
        nv_connections.add_connection(
            left=str(host),
            right=f'{local_port}@{host}',
        )
        nv_connections.add_connection(
            left=str(neighbour),
            right=f'{neighbour_port}@{neighbour}',
        )
        nv_connections.add_connection(
            left=f'{local_port}@{host}',
            right=f'{neighbour_port}@{neighbour}',
        )


def create_l2_topology(
        case: str,
        host_cache: HostCache,
        inv_columns: InventoryColumns,
        l2_drop_hosts: List[str],
        l2_host_map: Dict[str, str],
        l2_neighbour_replace_regex: List[Tuple[str, str]],
        label_: str,
        nv_connections: NvConnections,
        nv_objects: NvObjects,
        path_in_inventory: str,
        prefix: str,
        remove_domain: bool,
        seed_devices: Sequence[str],
) -> None:
    devices_to_go = list(set(seed_devices))  # remove duplicates
    devices_done = []

    while devices_to_go:
        device = devices_to_go[0]

        if device in l2_host_map.keys():
            try:
                devices_to_go.remove(device)
            except ValueError:
                pass
            device = l2_host_map[device]
            if device in devices_done:
                continue

        topo_data = host_cache.get_data(
            host=device, item=CacheItems.inventory, path=path_in_inventory
        )
        if topo_data:
            create_l2_device_from_inv(
                host=device,
                inv_data=topo_data,
                inv_columns=inv_columns,
                l2_host_map=l2_host_map,
                l2_drop_hosts=l2_drop_hosts,
                l2_neighbour_replace_regex=l2_neighbour_replace_regex,
                host_cache=host_cache,
                nv_objects=nv_objects,
                nv_connections=nv_connections,
                case=case,
                prefix=prefix,
                remove_domain=remove_domain
            )

            for _entry in nv_objects.host_list:
                if _entry not in devices_done:
                    devices_to_go.append(_entry)

        devices_to_go = list(set(devices_to_go))
        devices_done.append(device)
        devices_to_go.remove(device)
        LOGGER.info(msg=f'Device done: {device}, source: {label_}')


def create_l3v4_topology(
        emblems: Emblems,
        host_cache: HostCache,
        ignore_hosts: Sequence[str],
        ignore_ips: Sequence[IPv4Network],
        ignore_wildcard: Sequence[Wildcard],
        include_hosts: bool,
        nv_connections: NvConnections,
        nv_objects: NvObjects,
        replace: Mapping[str, str],
        skip_if: bool,
        skip_ip: bool,
        summarize: Sequence[IPv4Network],
) -> None:
    host_list: Sequence[str] = host_cache.get_hosts_by_label(HOST_LABEL_L3V4_ROUTER)

    if include_hosts:
        host_list += host_cache.get_hosts_by_label(HOST_LABEL_L3V4_HOSTS)

    LOGGER.debug(f'host list: {host_list}')
    if not host_list:
        LOGGER.warning(
            msg='No (routing capable) host found. Check if "inv_ip_addresses.mkp" '
                'added/enabled and inventory and host label discovery has run.'
        )
        return

    LOGGER.debug(f'L3v4 ignore hosts: {ignore_hosts}')
    for raw_host in host_list:
        host = raw_host
        if host in ignore_hosts:
            LOGGER.info(f'L3v4 host {host} ignored')
            continue
        if not (ipv4_addresses := host_cache.get_data(
            host=host, item=CacheItems.inventory, path=PATH_L3v4)
        ):
            LOGGER.warning(f'No IPv4 address inventory found for host: {host}')
            continue

        nv_objects.add_host(host=host, host_cache=host_cache)
        for _entry in ipv4_addresses:
            emblem = emblems.ip_network
            try:
                ipv4_info = Ipv4Info(**_entry)
            except TypeError:  # as e
                LOGGER.warning(f'Drop IPv4 address data for host: {host}, data: {_entry}')
                continue

            if ipv4_info.address.startswith('127.'):  # drop loopback addresses
                LOGGER.info(f'host: {host} dropped loopback address: {ipv4_info.address}')
                continue

            if ipv4_info.cidr == 32:  # drop host addresses
                LOGGER.info(
                    f'host: {host} dropped host address: {ipv4_info.address}/{ipv4_info.cidr}'
                )
                continue

            if ipv4_info.type.lower() != 'ipv4':  # drop if not ipv4
                LOGGER.warning(
                    f'host: {host} dropped non ipv4 address: {ipv4_info.address},'
                    f' type: {ipv4_info.type}'
                )
                continue

            if is_ignore_ipv4(ipv4_info.address, ignore_ips):
                LOGGER.info(f'host: {host} dropped ignore address: {ipv4_info.address}')
                continue

            if is_ignore_wildcard(ipv4_info.address, ignore_wildcard):
                LOGGER.info(f'host: {host} dropped wildcard address: {ipv4_info.address}')
                continue

            if network := get_network_summary(
                ipv4_address=ipv4_info.address,
                summarize=summarize,
            ):
                emblem = emblems.l3v4_summarize
                LOGGER.info(
                    f'Network summarized: {ipv4_info.network}/{ipv4_info.cidr} -> {network}'
                )
            else:
                network = f'{ipv4_info.network}/{ipv4_info.cidr}'

            if network in replace.keys():
                LOGGER.info(f'Replaced network {network} with {replace[network]}')
                network = replace[network]
                emblem = emblems.l3v4_replace

            nv_objects.add_ipv4_network(network=network, emblem=emblem)

            if skip_if is True and skip_ip is True:
                nv_connections.add_connection(left=host, right=network)
            elif skip_if is True and skip_ip is False:
                nv_objects.add_ipv4_address(
                    host=host,
                    interface=None,
                    ipv4_address=ipv4_info.address,
                    emblem=emblems.ip_address,
                )
                nv_objects.add_tooltip_quickinfo(
                    '{ipv4_info.address}@{host}', 'Interface', ipv4_info.device
                )
                nv_connections.add_connection(left=f'{host}', right=f'{ipv4_info.address}@{host}')
                nv_connections.add_connection(left=network, right=f'{ipv4_info.address}@{host}')
            elif skip_if is False and skip_ip is True:
                nv_objects.add_interface(
                    host=host, service=ipv4_info.device, host_cache=host_cache
                )
                nv_objects.add_tooltip_quickinfo(
                    f'{ipv4_info.device}@{host}', 'IP-address', ipv4_info.address
                )
                nv_connections.add_connection(left=f'{host}', right=f'{ipv4_info.device}@{host}')
                nv_connections.add_connection(left=network, right=f'{ipv4_info.device}@{host}')
            else:
                nv_objects.add_ipv4_address(
                    host=host,
                    interface=ipv4_info.device,
                    ipv4_address=ipv4_info.address,
                    emblem=emblems.ip_address,
                )
                nv_objects.add_interface(
                    host=host, service=ipv4_info.device, host_cache=host_cache,
                )
                nv_connections.add_connection(
                    left=host, right=f'{ipv4_info.device}@{host}')
                nv_connections.add_connection(
                    left=f'{ipv4_info.device}@{host}',
                    right=f'{ipv4_info.address}@{ipv4_info.device}@{host}',
                )
                nv_connections.add_connection(
                    left=network, right=f'{ipv4_info.address}@{ipv4_info.device}@{host}',
                )
