#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-12
# File  : nvdct/lib/utils.py

from collections.abc import Mapping, Sequence
from ast import literal_eval
from dataclasses import dataclass
from enum import Enum, unique
from json import dumps
from logging import disable as log_off, Formatter, getLogger, StreamHandler
from logging.handlers import RotatingFileHandler
from pathlib import Path
from re import match as re_match
from socket import socket, AF_UNIX, AF_INET, SOCK_STREAM, SHUT_WR
from sys import stdout, exit as sys_exit
from time import time as now_time
from tomllib import loads as toml_loads, TOMLDecodeError
from typing import List, Dict, TextIO

from lib.constants import (
    CMK_SITE_CONF,
    COLUMNS_CDP,
    COLUMNS_LLDP,
    HOST_LABEL_CDP,
    HOST_LABEL_L3V4_ROUTER,
    HOST_LABEL_LLDP,
    LABEL_CDP,
    LABEL_L3v4,
    LABEL_LLDP,
    LOGGER,
    OMD_ROOT,
    PATH_CDP,
    PATH_L3v4,
    PATH_LLDP,
    DATAPATH,
)

@unique
class ExitCodes(Enum):
    OK = 0
    BAD_OPTION_LIST = 1
    BAD_TOML_FORMAT = 2
    BACKEND_NOT_IMPLEMENTED = 3
    AUTOMATION_SECRET_NOT_FOUND = 4
    NO_LAYER_CONFIGURED = 5

@dataclass(frozen=True)
class Ipv4Info:
    address: str
    device: str
    broadcast: str
    cidr: int
    netmask: str
    network: str
    type: str

@dataclass(frozen=True)
class InventoryColumns:
    neighbour: str
    local_port: str
    neighbour_port: str

@dataclass(frozen=True)
class Layer:
    path: str
    columns: str
    label: str
    host_label: str

LAYERS = {
    'CDP': Layer(
        path=PATH_CDP,
        columns=COLUMNS_CDP,
        label=LABEL_CDP,
        host_label=HOST_LABEL_CDP,
    ),
    'LLDP': Layer(
        path=PATH_LLDP,
        columns=COLUMNS_LLDP,
        label=LABEL_LLDP,
        host_label=HOST_LABEL_LLDP,
    ),
    'L3v4': Layer(
        path=PATH_L3v4,
        columns='',
        label=LABEL_L3v4,
        host_label=HOST_LABEL_L3V4_ROUTER,
    ),
}


def get_local_cmk_version() -> str:
    return Path(f'{OMD_ROOT}/version').readlink().name


def get_local_cmk_api_port() -> int | None:
    data = get_data_from_toml(CMK_SITE_CONF)
    return data.get('CONFIG_APACHE_TCP_PORT')


def get_data_form_live_status(query: str) -> Dict | List | None:
    address = f'{OMD_ROOT}/tmp/run/live'
    # family = AF_INET if type(address) is tuple else AF_UNIX
    family = AF_INET if isinstance(address, tuple) else AF_UNIX
    sock = socket(family, SOCK_STREAM)
    sock.connect(address)
    sock.sendall(query.encode())
    sock.shutdown(SHUT_WR)
    chunks: List = []
    while len(chunks) == 0 or chunks[-1] != "":
        chunks.append(sock.recv(4096).decode())
    sock.close()
    if len(chunks) > 0:
        reply = literal_eval("".join(chunks).strip())
        if reply != [[b'']]:
            return reply
    return None


def get_data_from_toml(file: str) -> Dict:
    data = {}
    toml_file = Path(file)
    if toml_file.exists():
        try:
            data = toml_loads(toml_file.read_text())
        except TOMLDecodeError as e:
            LOGGER.exception(
                msg=f'ERROR: data file {toml_file} is not in valid TOML format! ({e}),'
                    f' (see https://toml.io/en/)'
            )
            sys_exit(ExitCodes.BAD_TOML_FORMAT.value)

    else:
        LOGGER.error(msg=f'WARNING: User data {file} not found.')
    LOGGER.info(msg=f'TOML file read: {file}')
    LOGGER.debug(msg=f'Data from toml file: {data}')
    return data


def rm_tree(root: Path) -> None:
    # safety
    if not str(root).startswith(DATAPATH):
        LOGGER.warning(msg=f"WARNING: bad path to remove, {str(root)}, don\'t delete it.")
        return
    for p in root.iterdir():
        if p.is_dir():
            rm_tree(p)
        else:
            p.unlink()
    root.rmdir()


def remove_old_data(keep: int, min_age: int, raw_path: str, protected: Sequence[str]) -> None:
    path: Path = Path(raw_path)
    default_topo = path.joinpath('default')
    directories = [str(directory) for directory in list(path.iterdir())]
    # keep default top
    if str(default_topo) in directories:
        directories.remove(str(default_topo))
        keep -= 1
        if default_topo.is_symlink():
            try:
                directories.remove(str(default_topo.readlink()))
            except ValueError:
                pass

    # keep protected topologies
    for directory in protected:
        try:
            directories.remove(str(path.joinpath(directory)))
        except ValueError:
            pass
        else:
            LOGGER.info(msg=f'Protected topology: {directory}, will not be deleted.')

    if len(directories) < keep < 1:
        return

    topo_by_age = {}

    for directory in directories:
        if Path(directory).is_dir():
            topo_by_age[Path(directory).stat().st_ctime] = directory

    topo_age = list(topo_by_age.keys())
    topo_age.sort()

    while len(topo_by_age) > keep:
        if min_age * 86400 > now_time() - topo_age[0]:
            LOGGER.info(
                msg=f'Topology "{Path(topo_by_age[topo_age[0]]).name}'
                    f'" not older then {min_age} day(s). not deleted.'
            )
            return
        LOGGER.info(f'delete old topology: {topo_by_age[topo_age[0]]}')
        rm_tree(Path(topo_by_age[topo_age[0]]))
        topo_by_age.pop(topo_age[0])
        topo_age.pop(0)


def save_data_to_file(data: Mapping, path: str, file: str, make_default: bool) -> None:
    """
    Save the data as json file.
    Args:
        data: the topology data
        path: the path were to save the dat
        file: the file name to save  data in
        make_default: if True, create the symlink "default" with path as target

    Returns:
        None
    """

    path_file = f'{path}/{file}'
    save_file = Path(f'{path_file}')
    save_file.parent.mkdir(exist_ok=True, parents=True)
    save_file.write_text(dumps(data))

    parent_path = Path(f'{path}').parent
    if not Path(f'{parent_path}/default').exists():
        make_default = True
    if make_default:
        Path(f'{parent_path}/default').unlink(missing_ok=True)
        Path(f'{parent_path}/default').symlink_to(target=Path(path), target_is_directory=True)


# CMK version 2.2.x format
def save_topology(
        data: dict,
        base_directory: str,
        output_directory: str,
        dont_compare: bool,
        make_default: bool,
        topology_file_name: str,
) -> None:
    path = f'{base_directory}/{output_directory}'

    def _save():
        save_data_to_file(
            data=data,
            path=path,
            file=topology_file_name,
            make_default=make_default,
        )

    if dont_compare:
        _save()
    else:
        if not is_equal_with_default(
                data=data,
                file=f'{base_directory}/default/{topology_file_name}'
        ):
            _save()
        else:
            LOGGER.warning(
                msg='Topology matches default topology, not saved! Use'
                    '"--dont-compare" to save identical topologies.'
            )


def is_mac_address(mac_address: str) -> bool:
    """
    Checks if mac_address is a valid MAC address.
    Will only accept MAC address in the form "AA:BB:CC:DD:EE:FF" (lower case is also ok).
    Args:
        mac_address: the MAC address to check

    Returns:
        True if mac_address is a valid MAC address
        False if mac_address not a valid MAC address
    """
    re_mac_pattern = '([0-9A-Z]{2}\\:){5}[0-9A-Z]{2}'
    if re_match(re_mac_pattern, mac_address.upper()):
        LOGGER.debug(msg=f'mac: {mac_address}, match')
        return True
    else:
        LOGGER.debug(msg=f'mac: {mac_address}, no match')
        return False


def is_list_of_str_equal(list1: List[str], list2: List[str]) -> bool:
    """
    Compares two list of strings. Before comparing the list will internal sorted.
    Args:
        list1:
        list2:

    Returns:
        True if both lists match
        False if they don't match
    """
    tmp_list1 = list1.copy()
    tmp_list2 = list2.copy()
    tmp_list1.sort()
    tmp_list2.sort()
    return tmp_list1 == tmp_list2


def is_valid_hostname(host: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,253}$'
    if re_match(re_host_pattern, host):
        return True
    else:
        LOGGER.error(f'Invalid hostname found: {host}')
        return False

def is_valid_site_name(site: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,16}$'
    if re_match(re_host_pattern, site):
        return True
    else:
        LOGGER.error(f'Invalid site name found: {site}')
        return False

def is_valid_customer_name(customer: str) -> bool:
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_]{1,16}$'
    if re_match(re_host_pattern, customer):
        return True
    else:
        LOGGER.error(f'Invalid customer name found: {customer}')
        return False


def is_valid_output_directory(directory: str) -> bool:
    # 2024-12-11T17:35:08.12
    re_host_pattern = r'^[0-9a-z-A-Z\.\-\_\:]{1,30}$'
    if re_match(re_host_pattern, directory):
        return True
    else:
        LOGGER.error(f'Invalid output directory name found: {directory}')
        return False

def is_valid_log_file(log_file: str) -> bool:
    if not log_file.startswith(f'{OMD_ROOT}/var/log/'):
        LOGGER.error(f'Logg file needs to be under "{OMD_ROOT}/var/log/"! Got {Path(log_file).absolute()}')
        return False
    return True


# not used in cmk 2.3.x format
def merge_topologies(topo_pri: Dict, topo_sec: Dict) -> Dict:
    """
    Merge dict_prim into dict_sec
    Args:
        topo_pri: data of dict_pri will overwrite the data in dict_sec
        topo_sec: dict where the data of dict_pri will be merged to

    Returns:
        Dict: topo_sec that contains merged data from top_sec and top_pri
    """
    keys_pri = list(topo_pri.keys())

    # first transfer all completely missing items from dict_prim to dict_sec
    for key in keys_pri:
        if key not in topo_sec.keys():
            topo_sec[key] = topo_pri[key]
        else:
            topo_sec[key]['connections'].update(topo_pri[key].get('connections', {}))
            topo_sec[key]['interfaces'] = list(
                set((topo_sec[key]['interfaces'] + topo_pri[key].get('interfaces', [])))
            )
        topo_pri.pop(key)
    return topo_sec


def compare_dicts(dict1: Mapping, dict2: Mapping) -> bool:
    # check top level keys
    if not is_list_of_str_equal(list(dict1.keys()), list(dict2.keys())):
        LOGGER.debug('Top level dont match')
        LOGGER.debug(f'dict1: {list(dict1.keys())}')
        LOGGER.debug(f'dict1: {list(dict2.keys())}')
        return False

    for key, value in dict1.items():
        _type = type(value)
        if _type == dict:
            if not compare_dicts(value, dict2[key]):
                return False
        elif _type == list:
            if not is_list_of_str_equal(value, dict2[key]):
                LOGGER.debug(f'list1: {value}')
                LOGGER.debug(f'list2: {dict2[key]}')
                return False
        elif _type == str:
            if not value == dict2[key]:
                LOGGER.debug('value dont match')
                LOGGER.debug(f'value1: {value}')
                LOGGER.debug(f'value2 {dict2[key]}')
                return False
        else:
            return False

    return True


def is_equal_with_default(data: Mapping, file: str) -> bool:
    default_file = Path(file)
    if default_file.exists():
        default_data = literal_eval(default_file.read_text())
        return compare_dicts(data, default_data)
    return False


def get_table_from_inventory(inventory: Dict[str, object], raw_path: str) -> List | None:
    path: List[str] = ('Nodes,' + ',Nodes,'.join(raw_path.split(',')) + ',Table,Rows').split(',')
    try:
        table = inventory.copy()
    except AttributeError:
        return None
    for m in path:
        try:
            # print(raw_table[m])
            table = table[m]
        except KeyError:
            LOGGER.info(msg=f'Inventory table for {path} not found')
            return None
    return list(table)


def configure_logger(log_file: str, log_level: int, log_to_console: bool) -> None:
    if log_level < 0:
        log_off()
        return

    log = getLogger()
    log_formatter = Formatter(
        fmt='%(asctime)s :: %(levelname)s :: %(module)s :: %(funcName)s() :: %(lineno)s :: %(message)s',
    )

    log.setLevel(log_level)

    # create a new file > 5 mb size
    log_handler_file = RotatingFileHandler(
        filename=log_file,
        mode='a',
        maxBytes=5 * 1024 * 1024,
        backupCount=5,
        # encoding=None,
        # delay=0
    )

    log_handler_file.setFormatter(log_formatter)
    # log_handler_file.setLevel(INFO)
    log.addHandler(log_handler_file)

    if log_to_console:
        log_handler_console = StreamHandler(stdout)
        log_handler_console.setFormatter(log_formatter)
        # log_handler_console.setLevel(INFO)
        log.addHandler(log_handler_console)


class StdoutQuiet:
    def __init__(self, term: TextIO = stdout, quiet: bool = False):
        # super().__init__()
        self.term = term
        self._quiet = quiet
        self._org_stdout = stdout

    def write(self, text):
        if not self._quiet:
            self.term.write(text)
        else:
            pass

    def flush(self):
        self._org_stdout.flush()
