#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2

# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2023-10-08
# File  : nvdct_data.py

# 2023-10-10: initial release
# 2023-10-16: added options  --keep-max  and --min-age
# 2023-10-17: removed -p, -c, -d (use the long options instead: --path-in-inventory, --data-source,
#             --inventory-columns)
#             changed option --keep-max to --keep
#             added option --check-user-data-only
#             added SEED_DEVICES to user data, option -s/--sed-devices is now optional
#             refactoring
# 2023-10-18: fixed --make-default not working
#             added protected topologies
#             changed get inventory data from file access to request via live status query
#             get interfaces from autochecks changed to live status query
# 2023-10-19: cleanup removed "all" unused functions
#             added option -u, --user-data-file to provide a user specific data file
# 2023-10-20: changed option -m/--make-default to -d/--default (-m is needed for -m/--merge
#             added -m/--merge option
# 2023-10-27: improved matching of port name from CDP/LLDP data to interface (item) name of service
#             improved internal data handling (HostCache)
# 2023-10-28: reworked handling of layers (see option -l/--layers)
#             added option -l/-layers {CDP,CUSTOM,LLDP,STATIC} [{CDP,CUSTOM,LLDP,STATIC} ...]
#             added section CUSTOM_LAYERS to user data file
#             removed option -m/--merge, included in option -l/--layers
#             removed option --lldp, included in option -l/--layers
#             removed option --data-source, now handled in CUSTOM_LAYERS
#             removed option --inventory-columns, now handled in CUSTOM_LAYERS
#             removed option --path-in-inventory, now handled in CUSTOM_LAYERS
#             removed lower limits for --keep and --min-age
# 2023-1ß-29: fixed missing path from custom layer in host prefetch
# 2023-11-16: added option -b/--backend [LIVESTATUS, FILESYSTEM],
#             LIVESTATUS is the default for performance reasons -> for now only local site
#             FILESYSTEM fetches the data directly form the inventory files -> use in distributed
#                        environments
# 2023-11-28: implemented MULTISITE as backend, via livestatus -> use in distributed environments,
#             may become the default
# 2023-11-30: added RESTAPI backend --> this is kind of slow (20 seconds compared to 0.5
#             for MULTISITE)
# 2023-12-06: do not add devices from static connections to seed devices --> be more transparent
#             added prefix option -> needs testing (THX to andreas doehler for the suggestion)
# 2023-12-21: adjusted for streamlined CDP/LLDP inventory format,
#             needs at least CDP 0.7.1-20231221 and LLDP 0.9.2-2023122
#             changed default backend from LIVESTATUS to MULTISITE
#             added logger (option --log-level)
#             removed option --debug
# 2023-12-22: added options --log-file and --log-to-stdout
# 2024-01-02: added new output format (option --new-format)
#             fixed LOGGER.warn --> LOGGER.warning
#             improved detection of missing hosts/interfaces
# 2024-01-03: fixed query in MULTISITE.exist
#             added option --quiet
#             renamed from create_topology_data to nvdct
#             added layer3 IPv4 topology (experimental)
# 2024-01-04: added SETTINGS section to nvdct.toml file -> defaults -> overridden by
#             toml -> overridden by cli
#             only seed devices are additive (toml + cli)
#             added L3v4 toml options: L3V4_IGNORE_HOSTS, L3V4_SUMMARIZE, L3V3_REPLACE
#             moved from ~/local/bin/topology_data to ~/local/bin/nvdct
# 2024-01-28: fixed -b MULTISITE if not in distributed environment (falling back to local)
#             fixed crash in create_l3v4_topology if no host was found
# 2024-01-29: added new output path for cmk 2.3
#             changed don't output cmk 2.2 format if --new-format set
# 2024-02-25: added images for L3 Topo (IP-Address and IP-Network)
#             icon source includes: https://www.flaticon.com/free-icons
#             added option to drop neighbours by regex
#             fixed no inventory in L3v4
# 2024-03-14: added metadata for connection (thickness, color)
#             thickness can be configured via toml (MAP_SPEED_TO_THICKNESS)
#             color 'red' is used for speed mismatch
#             made node emblem configurable (for ip-address and ip-network) via toml (EMBLEMS)
# 2024-03-16: moved args, utils, classes to lib
#             split classes in to backends, settings and classes
# 2024-03-17: changed speed mismatch from native speed to operational speed
#             added duplex mismatch ('yellow')
#             added native vlan mismatch ('orange')
# 2024-03-25: added --pre-fetch, this will speed up the RESTAPI backend performance
#             min inv_cdp_cahe version: 0.7.1-20240320 -> host label nvdct/has_cdp_neighbours
#             min inv_lldp_cache version: 0.9.3-20240320 -> host label nvdct/has_lldp_neighbours
# 2024-03-27: added option --api-port, defaults to 80
# 2024-03-28: changed restapi get_interface_data to use one call to fetch all data
# 2024-03-29: added L3V4_IGNORE_IP and L3V4_IGNORE_WILDCARD to toml file
# 2024-03-31: fixed: L3V3_REPLACE -> L3V4_REPLACE
#             made input from toml more bullet proof
# 2024-04-03: change HostCacheMultiSite to be child of HostCacheLiveStatus
# 2024-04-06: added node icon warning
# 2024-04-07: added css for connection
# 2024-04-13: added html tooltip for connection issues
# 2024-04-15: added emblems for l3v4_summary and l3v4_replace
# 2024-04-16: fixed html tooltip for speed mismatch (Speed: 100 MBit/s (expected: 1 GBit/s)WARN)
#             added option --skip-l3-if
# 2024-04-17: automatically detect local CMK version -> no need for --new-format option any more
#             automatically detect local CONFIG_APACHE_TCP_PORT -> no need for --api-port option any more  # noqa: E501
# 2024-04-18: added option --skip-l3-ip
#             added tooltip for --skip-l3-if and --skip-l3-ip
# 2024-04-30: automatically create @default if it doesn't exist
# 2024-05-07: fixed handling of ignore wildcard in L3v4 topology
# 2024-05-17: added support for Extreme interfaces in LLDP without stack id (1 -> 1:1)
# 2024-05-18: added support for NXOS interfaces in LLDP in short from (Eth1/1 -> Ethernet1/1)
#             added IP-Address/IP-Network as quickinfo
# 2024-05-18: fixed crash non empty neighbour port (ThX to andreas doehler)
# 2024-06-05: fixed interface index padding
# 2024-06-06: added interfcae detection alias+index
#             added interface detection description + index
# 2024-06-09: moved topologiy helpers to lib/topologies.py
#             moved (default) config file(s) to ./conf/
# 2024-06-14: added debug code for bad IPv4 address data
# 2024-06-17: fixed bad IPv4 address data (just drop it)
# 2024-09-23: incompatible replaced options --lowercase/--uppercase with --case LOWER|UPPER
#             changed version output from settings to argparse action
#             incompatible removed backend FILESYSTEM -> will fallback to MULTISITE
#             incompatible
#             removed support for CMK2.2.x file format (removed option --new-format)
# 2024-09-24: added site filter for multisite deployments (MULTISITE only), option --filter-sites
#               and SITES section in toml file
#             added customer filter for MSP deployments (MULTISITE only), option --filter-customers
#               and section CUSTOMERS in toml file
# 2024-11-16: added better logging for missing L2 data
# 2024-11-17: added L2_NEIGHBOUR_REPLACE_REGEX (ThX to Frankb@checkmk forum for the base idea)
#             incompatible removed DROP_HOST_REGEX -> use L2_NEIGHBOUR_REPLACE_REGEX instead
#             incompatible changed section names in TOML file to better distinguish between L2 and L3v4
#               HOST_MAP -> L2_HOST_MAP
#               DROP_HOSTS -> L2_DROP_HOSTS
#               SEED_DEVICES -> L2_SEED_DEVICES
#             incompatible removed option -s, --seed-devices from CLI -> use TOML section L2_SEED_DEVICES instead
#             incompatible changed the option keep-domain to remove-domain -> don't mess with neighbor names by default
# 2024-12-08: incompatible: changed hostlabel for L3v4 topology to nvdct/l3v4_topology
#                           needs at least inv_ip_address inv_ip_address-0.0.5-20241209.mkp
# 2024-12-09: added option --include-l3-hosts
#             added site filter for RESTAPI backend
#             enabled customer filter for MULTISITE backend
# 2024-12-10: refactoring: moved topology code to topologies, removed all global variables, created main() function
# 2024-12-11: incompatible: changed default layers to None -> use the CLI option -l CDP or the configfile instead
#             incompatible: reworked static topology -> can now be used for each service, host/service name has to be
#                           exactly like in CMK. See ~/local/bin/nvdct/conf/nfdct.toml
#             moved string constants to lib/constants.py
#
#             minimum ip address inventory plugin versions
#               - inv_lnx_if_ip-0.0.4-20241210.mkp
#               - inv_ip_address-0.0.6-20241210.mkp
#               - inv_win_if_ip-0.0.3-20241210.mkp

# creating topology data json from inventory data
#
# This script creates the topology data file needed for the Checkmk "network_visualization" plugin
# by Andreas Boesl/schnetz. For more information see
# https://forum.checkmk.com/t/network-visualization/41680  (from 2023-10-05)
# https://exchange.checkmk.com/p/network-visualization     (from 2023-10-05)
#
# NOTE: the topology_data configuration (layout etc.) is saved under ~/var/check_mk/topology
#
# The inventory data could be created with my CDP/LLDP/IP Address/Interface nane inventory plugins:
# CDP.....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_cdp_cache
# LLDP....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lldp_cache
# L3v4....: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ip_addresses
#         : https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_lnx_if_ip
#         : https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_win_if_ip
# IF Name.: https://thl-cmk.hopto.org/gitlab/checkmk/vendor-independent/inventory/inv_ifname
#
# USAGE:
# ~/local/lib/nvdct/nvdct.py --help
#

"""
# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################

__objects = {
    "HostA": {
        "name": "myhost",
        "link": {"core": "myhost"},
        'metadata': {},
    },
    "Service@HostA": {
        "name": "SomeService",
        "link": {"core": ["myhost", "interface3"]},  # service@host in CMK
        # "link": {"core": "myhost"},                # host object in CMK
        # "link": {},                                # object not in CMK core
        'metadata': {
            'images': {
                'icon': 'icon_warning',              # additional image shown at the node
                'emblem': 'ip-network_80',           # image to show for an object
            },
            'tooltip': {
                'html': "Test message<div style='color: blue'>Some color <b>test</b></div>",
                'quickinfo': [
                                {'name': "Duplex", 'value': 'full duplex'},
                                {'name': 'Speed', 'value': '123132234'},
                                {'name': 'Native vlan', 'value': '123'}
                            ]
            }
        },
    },
}

__connections = [
    [["HostA", "ServiceOfHostA"], {}],
    [["ServiceOfHostA", "ServiceOfHostB"], {
        'line_config': {
            'thickness': 3,
            'color': '332211',
            'css_styles': {
                'stroke': 'green',
                'stroke-dasharray': '30 30',
                # 'stroke-dasharray': 'unset',  # to remove a CSS attribute
                'stroke-width': '4',
            }
        },
        'tooltip': {
            'html': "Test message mismatch<div style='color: blue'>Some color <b>test</b></div>",
            'quickinfo': [
                            {'name': "Duplex", 'value': 'full duplex'},
                            {'name': 'Native speed', 'value': '123132234'},
                            {'name': 'Native vlan', 'value': '123'}
                        ]
        }
    }],
    [["ServiceOfHostB", "HostB"], {}],
]

__data = {
    'version': 1,
    'name': 'LLDP',  # label for this layer
    'objects': __objects,
    'connections': __connections,
}

# ##################################################################################################
# ##################################################################################################
# ##################################### data format ################################################
# ##################################################################################################
# ##################################################################################################
"""

import sys
from logging import DEBUG
from time import strftime, time_ns
from typing import List

from lib.args import parse_arguments
from lib.backends import (
    HostCache,
    HostCacheLiveStatus,
    HostCacheMultiSite,
    HostCacheRestApi,
)
from lib.constants import (
    HOME_URL,
    LOGGER,
    NVDCT_VERSION,
    DATAPATH,
)
from lib.settings import Settings
from lib.topologies import (
    NvConnections,
    NvObjects,
    create_l2_topology,
    create_l3v4_topology,
    create_static_connections,
)
from lib.utils import (
    ExitCodes,
    InventoryColumns,
    LAYERS,
    Layer,
    StdoutQuiet,
    configure_logger,
    remove_old_data,
    save_data_to_file,
)


def main():
    start_time = time_ns()

    nv_connections = NvConnections()
    nv_objects = NvObjects()

    settings: Settings = Settings(vars(parse_arguments()))

    sys.stdout = StdoutQuiet(quiet=settings.quiet)
    configure_logger(
        log_file=settings.log_file,
        log_to_console=settings.log_to_stdtout,
        log_level=settings.loglevel,
    )
    # always logg start and end of a session (except --log-level OFF)
    LOGGER.critical(msg='Data creation started')

    print('')
    print(
        f'Network Visualisation Data Creation Tool (NVDCT)\n'
        f'by thl-cmk[at]outlook[dot]com, version {NVDCT_VERSION}\n'
        f'see {HOME_URL}'
    )
    print('')
    print(f'Start time....: {strftime(settings.time_format)}')

    match settings.backend:
        case 'RESTAPI':
            host_cache: HostCache = HostCacheRestApi(
                pre_fetch=settings.pre_fetch,
                api_port=settings.api_port,
                filter_sites=settings.filter_sites,
                sites=settings.sites
            )
        case 'MULTISITE':
            host_cache: HostCache = HostCacheMultiSite(
                pre_fetch=settings.pre_fetch,
                filter_sites=settings.filter_sites,
                sites=settings.sites,
                filter_customers=settings.filter_customers,
                customers=settings.customers,
            )
        case 'LIVESTATUS':
            host_cache: HostCache = HostCacheLiveStatus(
                pre_fetch=settings.pre_fetch,
            )
        case _:
            LOGGER.error(msg=f'Backend {settings.backend} not (yet) implemented')
            host_cache: HostCache | None = None  # to keep linter happy
            sys.exit(ExitCodes.BACKEND_NOT_IMPLEMENTED.value)

    jobs: List[Layer] = []
    pre_fetch_layers: List[str] = []
    pre_fetch_host_list: List[str] = []

    for layer in settings.layers:
        if layer == 'STATIC':
            jobs.append(layer)
        if layer == 'L3v4':
            jobs.append(layer)
            host_cache.add_inventory_prefetch_path(path=LAYERS[layer].path)
            pre_fetch_layers.append(LAYERS[layer].host_label)

        elif layer in LAYERS:
            jobs.append(LAYERS[layer])
            host_cache.add_inventory_prefetch_path(path=LAYERS[layer].path)
            pre_fetch_layers.append(LAYERS[layer].host_label)

        elif layer == 'CUSTOM':
            for entry in settings.custom_layers:
                jobs.append(entry)
                host_cache.add_inventory_prefetch_path(entry.path)

    if not jobs:
        message = ('No layer to work on. Please configura at least one layer (i.e. CLI option "-l CDP")\n'
                   'See ~/local/bin/nvdct/conf/nvdct.toml -> SETTINGS -> layers')
        LOGGER.warning(message)
        print(message)
        sys.exit(ExitCodes.NO_LAYER_CONFIGURED.value)

    if settings.pre_fetch:
        LOGGER.info('Pre fill cache...')
        for host_label in pre_fetch_layers:
            if _host_list := host_cache.get_hosts_by_label(host_label):
                pre_fetch_host_list = list(set(pre_fetch_host_list + _host_list))
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts start')
        print(f'Prefetch start: {strftime(settings.time_format)}')
        print(f'Prefetch hosts: {len(pre_fetch_host_list)} of {len(host_cache.cache.keys())}')
        host_cache.pre_fetch_cache(pre_fetch_host_list)
        LOGGER.info(f'Fetching data for {len(pre_fetch_host_list)} hosts end')
        print(f'Prefetch end..: {strftime(settings.time_format)}')

    for job in jobs:
        match job:
            case 'STATIC':
                label = 'STATIC'
                create_static_connections(
                    connections_=settings.static_connections,
                    emblems=settings.emblems,
                    host_cache=host_cache,
                    nv_objects=nv_objects,
                    nv_connections=nv_connections,
                )
            case 'L3v4':
                label = 'L3v4'
                create_l3v4_topology(
                    ignore_hosts=settings.l3v4_ignore_hosts,
                    ignore_ips=settings.l3v4_ignore_ips,
                    ignore_wildcard=settings.l3v4_ignore_wildcard,
                    include_hosts=settings.include_l3_hosts,
                    replace=settings.l3v4_replace,
                    skip_if=settings.skip_l3_if,
                    skip_ip=settings.skip_l3_ip,
                    summarize=settings.l3v4_summarize,
                    emblems=settings.emblems,
                    host_cache=host_cache,
                    nv_objects=nv_objects,
                    nv_connections=nv_connections,
                )
            case _:
                label = job.label.upper()
                columns = job.columns.split(',')
                create_l2_topology(
                    seed_devices=settings.l2_seed_devices,
                    path_in_inventory=job.path,
                    inv_columns=InventoryColumns(
                        neighbour=columns[0],
                        local_port=columns[1],
                        neighbour_port=columns[2]
                    ),
                    label_=label,
                    l2_drop_hosts=settings.l2_drop_hosts,
                    l2_host_map=settings.l2_host_map,
                    l2_neighbour_replace_regex=settings.l2_neighbour_replace_regex,
                    host_cache=host_cache,
                    nv_objects=nv_objects,
                    nv_connections=nv_connections,
                    case=settings.case,
                    prefix=settings.prefix,
                    remove_domain=settings.remove_domain,
                )

        nv_connections.add_meta_data_to_connections(
            nv_objects=nv_objects,
            speed_map=settings.map_speed_to_thickness,
        )

        _data = {
            'version': 1,
            'name': label,
            'objects': nv_objects.nv_objects if not settings.loglevel == DEBUG else dict(
                sorted(nv_objects.nv_objects.items())
            ),
            'connections': nv_connections.nv_connections if not settings.loglevel == DEBUG else sorted(
                nv_connections.nv_connections
            )
        }
        save_data_to_file(
            data=_data,
            path=(
                f'{DATAPATH}/{settings.output_directory}'
            ),
            file=f'data_{label}.json',
            make_default=settings.default,
        )

        message = (
            f'Layer {label:.<8s}: Devices/Objects/Connections added {nv_objects.host_count}/'
            f'{len(nv_objects.nv_objects)}/{len(nv_connections.nv_connections)}'
        )
        LOGGER.info(msg=message)
        print(message)

        nv_objects = NvObjects()
        nv_connections = NvConnections()

    if settings.keep:
        remove_old_data(
            keep=settings.keep,
            min_age=settings.min_age,
            raw_path=DATAPATH,
            protected=settings.protected_topologies,
        )

    print(f'Time taken....: {(time_ns() - start_time) / 1e9}/s')
    print(f'End time......: {strftime(settings.time_format)}')
    print('')

    LOGGER.critical('Data creation finished')


if __name__ == '__main__':
    main()
